#include "hermite.h"






//----------------------------------------------------------------------
hermite::hermite(interval support_)
{
  support = support_;
  support_g = support.inf();
  support_d = support.sup();
  milieu = 0.5*(support_g + support_d);
  
  expression temp = expression("((x0-x)^2*(-2*x-x0+3*x1))/((x1-x0)^3)");
  expression temp2 = temp.substitute(support.inf(), "x0");
  f = temp2.substitute(support.sup(), "x1");


  expression temp_prime = expression("(-6*(x0-x)*(x1-x))/((x1-x0)^3)");
  expression temp2_prime = temp_prime.substitute(support.inf(), "x0");
  f_prime = temp2_prime.substitute(support.sup(), "x1");
}




interval hermite::evalue(interval x)
{

  domain G;
  G["x"] = interval(x.inf());
  interval result_G = f.Evalue(G);

  domain D;
  D["x"] = interval(x.sup());
  interval result_D = f.Evalue(D);

  
  double result_g;
  double result_d;

  double g = x.inf();
  double d = x.sup();

  
  if (g < support_g)
    {
      result_g = 0;
    };

  if ((support_g <= g) && ( g <= support_d))
    {
      domain temp;
      result_g = result_G.inf();
    };

  if ( support_d < g)
    {
      domain temp;
      result_g = 1;
    };

  
  if (d < support_g)
    {
      result_d = 0;
    };
  
  if ((support_g <= d) && ( d <= support_d))
    {
      domain temp;
      result_d = result_D.sup();
    };
  
  if ( support_d < d)
    {
      domain temp;
      result_d = 1;
    };
  
  interval result(result_g,result_d);   
  return result;
}




interval hermite::evalue_prime(interval x)
{

  domain G;
  G["x"] = interval(x.inf());
  interval result_G = f_prime.Evalue(G);

  domain M;
  M["x"] = interval(milieu);
  interval result_M = f_prime.Evalue(M);
  
  domain D;
  D["x"] = interval(x.sup());
  interval result_D = f_prime.Evalue(D);
  
  double result_g;
  double result_d;

  double g = x.inf();
  double d = x.sup();

  
  if (g < support_g)
    {
      result_g = 0;
      if (d < support_g)
	{
	  result_d = 0;
	}
      else if (d < milieu)
	{
	  result_d = result_D.sup();
	}
      else
	{
	  result_d = result_M.sup();
	}
    }
  else if (g < milieu)
    {
      if (d < milieu)
	{
	  result_g = result_G.inf();
	  result_d = result_D.sup();
	}
      else if (d < support_d)
	{
       	  result_g = min(result_G.inf(), result_D.inf()) ;
	  result_d = result_M.sup();
	}
      else
	{
       	  result_g = 0;
	  result_d = result_M.sup();
	}
    }
  else if (g < support_d)
    {
      if (d <  support_d)
	{
	  result_g = result_D.inf();
	  result_d = result_G.sup();	 
	}
      else
	{
	  result_g = 0;
	  result_d = result_G.sup();	 
	}
    }
  else
    {
	  result_g = 0;
	  result_d = 0;
    }
      
    
  interval result(result_g,result_d);   
  return result;
}



/*
 _            _      __                  _   _             
| |_ ___  ___| |_   / _|_   _ _ __   ___| |_(_) ___  _ __  
| __/ _ \/ __| __| | |_| | | | '_ \ / __| __| |/ _ \| '_ \ 
| ||  __/\__ \ |_  |  _| |_| | | | | (__| |_| | (_) | | | |
 \__\___||___/\__| |_|  \__,_|_| |_|\___|\__|_|\___/|_| |_|
                                                           
*/

fonction_test::fonction_test(domain support_)
{
  support = support_;
  varphi_t = new hermite(support["t"]);
  varphi_x = new hermite(support["x"]);
}


interval fonction_test::evalue(domain entree)
{
  interval result;
  result = varphi_t->evalue(entree["t"]) * varphi_x->evalue(entree["x"]);
  return result;
}


interval fonction_test::evalue_d_dt(domain entree)
{
  interval result;
  result = varphi_t->evalue_prime(entree["t"]) * varphi_x->evalue(entree["x"]);
  return result;
}


interval fonction_test::evalue_d_dx(domain entree)
{
  interval result;
  result = varphi_t->evalue(entree["t"]) * varphi_x->evalue_prime(entree["x"]);
  return result;
}

 



void  fonction_test::to_gnuplot(char * filename)
{

  domain X;
  X["x"] = interval(-1,1);
  X["t"] = interval(-1,1);
  
  fstream sortie_gp;
  //  char *sPhrase2;
  sortie_gp.open(filename,fstream::out);
  sortie_gp << "set samples 21, 21 \n " << endl;
  sortie_gp << "set isosamples 11, 11 " << endl; 
  sortie_gp << "set title \"essai\"  " << endl; 
  sortie_gp << "set xlabel \"x axis\"  " << endl; 
  sortie_gp << "set xrange [ -1.0 : 1.0 ] noreverse nowriteback " << endl; 
  sortie_gp << "set ylabel \"t axis\"  " << endl;  
  sortie_gp << "set yrange [ -1.0 : 1.0 ] noreverse nowriteback " << endl; 
  sortie_gp << "set zlabel \"value axis\"  " << endl; 
  sortie_gp << "splot \"data.dat\" with lines " << endl; 
  sortie_gp << "pause -1 " << endl; 
  sortie_gp << "" << endl;
  sortie_gp.close();
  
  double pas_x = 0.01 * X["x"].width()  ;
  double pas_t = 0.01 * X["t"].width()  ;

  fstream data;
  data.open("data.dat",fstream::out);
  
  for ( double x = X["x"].inf() ; x < X["x"].sup() ; x = x + pas_x )
    {
    for ( double t = X["t"].inf() ; t < X["t"].sup() ; t = t + pas_t )
      {
	domain temp;
	temp["x"] = interval(x);
	temp["t"] = interval(t);
	//interval valeur = (*this).evalue_d_dx(temp);
	interval valeur = (*this).evalue(temp);
	data << x << " " << t  << " "  << valeur.inf() << endl;	
      }
    data << endl;
    }
  data.close();
}






