#include <iostream>
#include <string>
#include <stdlib.h>
#include <sstream>
#include <fstream>

#include <thread>

#include "vinter.h"
#include "hermite.h"
#include <glpk.h>             /* GNU GLPK linear/mixed integer solver */

#define card_var 300000

//---------------------------------------------------------------------
vinter::vinter (Map _f  , expression _h, expression _H, domain _K  , domain _T, domain _X , domain _U , domain _x_0)
{
  f = _f;
  h = _h;
  H = _H;
  T = _T;
  X = _X;
  U = _U;
  x_0 = _x_0;
  index_contrainte = 1;
  index_a = 1;
}

void vinter::construit_mesh(int nb)
{ 
  taille = nb;
  // Création du maillage pour mu
  domain total;
  total["x"] = X["x"];
  total["u"] = U["u"];
  mu = mesh(total,nb);

  // Création du maillage pour nu  
  domain total_final;
  total_final["x"] = X["x"];
  nu = mesh(total_final,1*nb);
};


void vinter::construit_cout()
{
  int mu_index = 1;
  int nu_index = 1;
  int index_variables = 1;

  lp = glp_create_prob();
  glp_set_prob_name(lp, "primal_relache");
  glp_set_obj_dir(lp, GLP_MIN);

  glp_add_cols(lp, mu.size() + nu.size());
    
  for (mesh::const_iterator i = mu.begin(); i!=mu.end();++i)
    {
      interval cost = h.Evalue(*i);
      std::string var_mui = "mu";
      var_mui += std::to_string(mu_index);
      glp_set_col_name(lp, index_variables, var_mui.c_str());
      glp_set_col_bnds(lp, index_variables, GLP_LO, 0.0, 0.0);
      glp_set_obj_coef(lp, index_variables, cost.inf());
      mu_index++;
      index_variables++;
    }

  for (mesh::const_iterator i = nu.begin(); i!=nu.end();++i)
    {
      interval cost = H.Evalue(*i);
      std::string var_nui = "nu";
      var_nui += std::to_string(nu_index);
      glp_set_col_name(lp, index_variables, var_nui.c_str());
      glp_set_col_bnds(lp, index_variables, GLP_LO, 0.0, 0.0);
      glp_set_obj_coef(lp, index_variables, cost.inf());
      nu_index++;
      index_variables++;
    }  
};



void vinter::contraintes_temporelles()
{
  int nombre_variable = taille * taille + taille;
  int *ind_inf;
  ind_inf = new int[nombre_variable + 1];  
  double *val_inf;
  val_inf = new double[nombre_variable + 1];
  int *ind_sup;
  ind_sup = new int[nombre_variable + 1];  
  double *val_sup;
  val_sup = new double[nombre_variable + 1];
   
  int index_variable = 1;
  for (mesh::const_iterator i = mu.begin(); i!=mu.end();++i)
    {
      domain x_i = *i;
      ind_inf[index_variable] = index_variable;
      val_inf[index_variable] = 1.0;

      ind_sup[index_variable] = index_variable;
      val_sup[index_variable] = 1.0;
      
      index_variable++;
    }

  for (mesh::const_iterator j = nu.begin(); j!=nu.end();++j)
    {
      domain y_j = *j;

      ind_inf[index_variable] = index_variable;
      val_inf[index_variable] = 0.0;

      ind_sup[index_variable] = index_variable;
      val_sup[index_variable] = 0.0;

      index_variable++;
    }

  glp_add_rows(lp, 1);
  string debut; debut = "temporelle_i"; 
  glp_set_row_name(lp, index_contrainte , debut.c_str());
  glp_set_row_bnds(lp, index_contrainte, GLP_UP, 0, 2.0 );
  glp_set_mat_row(lp, index_contrainte, index_variable -1, ind_inf, val_inf);

  index_contrainte++;

  glp_add_rows(lp, 1);
  debut = "temporelle_s";
  glp_set_row_name(lp, index_contrainte , debut.c_str());
  glp_set_row_bnds(lp, index_contrainte, GLP_LO, 2.0 , 0);
  glp_set_mat_row(lp, index_contrainte, index_variable -1, ind_sup, val_sup);
  index_contrainte++;
  
};



/*
void vinter::contraintes_temporelles()
{
  
  // l'idée est de collecter les pavés qui sont parallèles temporellement ensemble
  // afin d'en faire une contrainte
   
  // On ajoute "taille" contraintes
  // Chacune des contraintes correspond à une tranche temporelle
  glp_add_rows(lp, taille);
  index_contrainte = index_contrainte + taille;
  
  //  la case cons[i] contient un vecteur de paire (indice_variable,coef)
  vector <  vector < pair < int, double > > > cons;
  for (int  i = 1; i <= taille ; i++)
    { 
      string debut;
      debut = "contrainte_tempo_" +  std::to_string(i) ;
      interval pas_temporel = interval(T["t"].sup()) / interval ((double) taille, (double) taille ) ;
      glp_set_row_name(lp, i, debut.c_str());
      glp_set_row_bnds(lp, i, GLP_DB, pas_temporel.inf()-0.00000001 , pas_temporel.sup() + 0.0000001 );
      vector < pair < int, double > > vide;
      cons.push_back(vide);
    }

  // on ajoute les paires à la bonne contrainte
  int mu_index = 1;
  for (mesh::const_iterator p = mu.begin(); p!=mu.end();++p)
    {
      domain temp_p = *p;
      int contrainte_i = (int) floor ( (taille * temp_p["t"].mid()) / T["t"].sup()) ;      
      pair < int, double > ma_paire(mu_index, 1.0);
      //      cout << mu_index  << "," << contrainte_i << endl; 
      cons[contrainte_i].push_back(ma_paire);
      mu_index++;
    }

  
  // on ajoute les contraintes à l'objet glp_prob
  for (int  i = 0; i < cons.size() ; i++)
    {
      int nombre_variable = taille * taille * taille ;
      int *les_indices;
      les_indices = new int[nombre_variable + 1];
      
      double *les_valeurs;
      les_valeurs = new double[nombre_variable + 1];
      
      for (int b = 0; b < cons[i].size() ; b++)
	{
	  les_indices[b+1] = cons[i][b].first;
	  les_valeurs[b+1] = cons[i][b].second;
	};
      glp_set_mat_row(lp, i+1,  cons[i].size(), les_indices, les_valeurs);
      
      delete[] les_indices;
      delete[] les_valeurs;
    }
};
*/

void vinter::write()
{
  glp_write_lp(lp, NULL , "test.lp");
}
  

void vinter::solution_to_gnuplot(char * file)
{

      ofstream fichier(file);      
      int mu_index = 1;
      int nu_index = 1;
      int index_variables = 1; 

      for (mesh::const_iterator i = mu.begin(); i!=mu.end();++i)
	{
	  domain p = *i;
	  double zj = glp_get_col_prim(lp, index_variables);	  
	  fichier << zj << " " ;
	  if ( (index_variables%taille)==0)
	    {
	      fichier  << endl ;
	    }
	  index_variables++;
	}
      fichier.close();
};






double vinter::solve(string file)
{
  //  glp_exact(lp, NULL);
    glp_simplex(lp, NULL);


    //  glp_interior(lp,NULL);

  //  glp_write_lp(lp, NULL , "test.lp");
  /*glp_write_mps(lp, GLP_MPS_FILE, NULL , "test_mps.lp");
  */



    // Pour le dessin des solutions en 3D
  // ------------------------------------------------------------
      ofstream fichier("output/"+file); 
      fichier << "object{ Superboxvide(" << 10*T["t"].inf() << "," <<  10*X["x"].inf() << "," <<  4*U["u"].inf() 
	      << "," <<  10*T["t"].sup() << "," <<  10*X["x"].sup() << "," <<  4*U["u"].sup() 
	      << "," << 1 << "," << "0.01"  << " ) }" << endl;
      
      int mu_index = 1;
      int nu_index = 1;
      int index_variables = 1; 
      
      for (mesh::const_iterator i = mu.begin(); i!=mu.end();++i)
	{
	  domain p = *i;
	  double zj = glp_get_col_prim(lp, index_variables);	  
	  double diam = 0.01;
       	  if ( zj > 0.0001 ) 
	    {
	      fichier << "object{ Superbox(" << 10*p["t"].inf() << "," <<  10*p["x"].inf() << "," <<  4*p["u"].inf() 
		      << "," <<  10*p["t"].sup() << "," <<  10*p["x"].sup() << "," <<  4*p["u"].sup() 
		      << "," << 1-zj << "," << diam  << " ) }" << endl;
	    }
	  index_variables++;
	}

      fichier << "// ici nu ---------" << endl;
      
      for (mesh::const_iterator i = nu.begin(); i!=nu.end();++i)
	{
	  domain p = *i;
	  double zj = glp_get_col_prim(lp, index_variables);	  
	  //	  printf("pour nu = %f; \n", zj);
	  double diam = 0.02;
       	  if ( zj > 0.0002 ) 
	    {
	      fichier << "object{ Superbox(" << 10*p["t"].inf()-0.01 << "," <<  10*p["x"].inf() << "," <<  4*U["u"].inf()-0.01
		      << "," <<  10*p["t"].sup()+0.1 << "," <<  10*p["x"].sup() << "," <<  4*U["u"].inf()+0.01
		      << "," << 1-zj << "," << diam  << " ) }" << endl;
	    }
	  index_variables++;
	}
      fichier.close();
  // Fin du dessin en 3D
  // ------------------------------------------------------------

      double z = glp_get_obj_val(lp);

   return z;
};


int  vinter::cardinal_variable()
{
  int nombre_variable = glp_get_num_cols(lp);
  return nombre_variable;
}
;


int  vinter::cardinal_contraints()
{  
  int nombre_contraintes = glp_get_num_rows(lp);
  return nombre_contraintes;
};




void vinter::show_result()
{
       int mu_index = 1;
       int nu_index = 1;
       int index_variables = 1; 
      
 
  cout << "// ici mu ---------" << endl;      
  for (mesh::const_iterator i = mu.begin(); i!=mu.end();++i)
    {
      domain p = *i;
      p.out();
	  //p.out();
	  double zj = glp_get_col_prim(lp, index_variables);	  
	  //       	  cout << "solution : " <<  zj << endl;
       	  cout <<  zj << endl;
	  
	  index_variables++;
	}

      cout << "// ici nu ---------" << endl;      
      for (mesh::const_iterator i = nu.begin(); i!=nu.end();++i)
	{
	  domain p = *i;
	  p.out();
	  double zj = glp_get_col_prim(lp, index_variables);	  
	  cout << zj << endl;
	  index_variables++;
	}

}




void vinter::construit_contraintes(int debut, int fin )
{

  for (int deg_x = 1; deg_x < fin ; deg_x++  )
    {
      std::ostringstream monome;
      monome << "x^" << deg_x ;
      std::ostringstream nom_c;
      nom_c << "x^" << deg_x ;
      contrainte( monome.str() , nom_c.str());
    }
  
};



double vinter::test_vrai_solution()
{
  double lb_to_optimal_value = 0;
  ofstream fichier("output/vrai.pov");
  
  fichier << "object{ Superboxvide(" << 5*T["t"].inf() << "," <<  X["x"].inf() << "," <<  3*U["u"].inf() 
	  << "," <<  5*T["t"].sup() << "," <<  X["x"].sup() << "," <<  3*U["u"].sup() 
	  << "," << 1 << "," << "0.01"  << " ) }" << endl;
  
  mesh temporelle(T,taille);
  cout << endl;
  for (mesh::const_iterator i = temporelle.begin(); i!=temporelle.end();++i)
    {
      domain t  = *i;
      //      t.center().out();
      domain solution = t.center();

      //    cout << log(2) << endl;
      if (t["t"].inf() < log(2))
	{
	  solution["u"] = interval(0.0);
	  double x =  exp(t["t"].inf()) - 1;
	  solution["x"] = interval(x,x);	 
	}
      else
	if (t["t"].inf() < 2 - log(2))
	  {
	    solution["u"] = interval(2.0);
	    solution["x"] = interval(1.0);
	  }
	else
	  {
	    solution["u"] = interval(0.0);
	    double tau = t["t"].inf();
	    double x =  4*exp(-2+tau) - 1;
	    solution["x"] = interval(x,x);	 
	  };

      int lacase = mu.which_one(solution.center());
      domain p = mu[lacase];

      double zj = 1;	  
      double diam = 0.01;

      fichier << "object{ Superbox(" << 5*p["t"].inf() << "," <<  p["x"].inf() << "," <<  3*p["u"].inf() 
	      << "," <<  5*p["t"].sup() << "," <<  p["x"].sup() << "," <<  3*p["u"].sup() 
	      << "," << 1-zj << "," << diam  << " ) }" << endl;
      

      lb_to_optimal_value = t["t"].diam()*h.Evalue(p).inf() + lb_to_optimal_value ;	   
    }
  fichier.close();
  return lb_to_optimal_value ;
};


string  vinter::contraintes_inclusion()
{
  return "";
}



string vinter::dirac_final()
{

  int ind[card_var];
  double val[card_var];
 
  glp_add_rows(lp, 1);

  glp_set_row_name(lp, index_contrainte, "dirac");
  glp_set_row_bnds(lp, index_contrainte, GLP_DB, 1-0.00000000001, 1+0.00000000001);
 
  int compteur = 1;
  int nu_index = mu.size() + 1; // nu commence à la fin de mu
  for (mesh::const_iterator p = nu.begin(); p!=nu.end();++p)
    {
      ind[compteur] = nu_index;
      val[compteur] = 1;
      nu_index++;
      compteur++;
    }

  glp_set_mat_row(lp, index_contrainte , compteur - 1, ind, val); 
  index_contrainte = index_contrainte + 1;
  return "";
}

void vinter::contrainte(string polynome ,  string nom)
{

  int nombre_variable = taille * taille * taille + 1 * taille;
  int *ind_inf;
  ind_inf = new int[nombre_variable + 1];  
  double *val_inf;
  val_inf = new double[nombre_variable + 1];
  int *ind_sup;
  ind_sup = new int[nombre_variable + 1];  
  double *val_sup;
  val_sup = new double[nombre_variable + 1];
  
  mtx.lock();
  expression varphi(polynome);
  expression psi = expression("-1")* ( varphi.derive("t")  + varphi.derive("x") * f.component["x"]);
  cout << varphi.chaine  << endl;
  cout << psi.chaine << endl;
  mtx.unlock();


  int index_variable = 1;
  for (mesh::const_iterator i = mu.begin(); i!=mu.end();++i)
    {
      domain x_i = *i;
      //      double lb_psi = psi.Evalue(x_i).inf();
      interval lb_up_psi = psi.Evalue_en_coupant(x_i,4);
      ind_inf[index_variable] = index_variable;
      val_inf[index_variable] = lb_up_psi.inf();

      //      double up_psi = psi.Evalue(x_i).sup();
      ind_sup[index_variable] = index_variable;
      val_sup[index_variable] = lb_up_psi.sup();
      
      /*
      cout << "--------------------------" << endl;
      cout << index_variable << endl ;
      x_i.out();
      cout << lb_psi << endl;
      cout << up_psi << endl;
      */
      index_variable++;
    }

  for (mesh::const_iterator j = nu.begin(); j!=nu.end();++j)
    {
      domain y_j = *j;
      //      temp.out();
      double lb_varphi = varphi.Evalue(y_j).inf();

      ind_inf[index_variable] = index_variable;
      val_inf[index_variable] = lb_varphi;


      double up_varphi = varphi.Evalue(y_j).sup();
      ind_sup[index_variable] = index_variable;
      val_sup[index_variable] = up_varphi;
      /*
      cout << "--------------------------" << endl;
      cout << index_variable << endl ;
      y_j.out();
      cout << lb_varphi << endl;
      cout << up_varphi << endl;
      */

      index_variable++;
    }

  /*  domain x_0_etendue = x_0;
  x_0_etendue["t"] = interval(T["t"].inf(), T["t"].inf());
  */
  
  mtx.lock();
  glp_add_rows(lp, 1);
  string debut; debut = "contrainte_i_" + nom; 
  glp_set_row_name(lp, index_contrainte , debut.c_str());
  glp_set_row_bnds(lp, index_contrainte, GLP_UP, 0, varphi.Evalue(x_0).inf() );
  glp_set_mat_row(lp, index_contrainte, index_variable -1, ind_inf, val_inf);

  index_contrainte++;

  glp_add_rows(lp, 1);
  debut = "contrainte_s_" + nom;
  glp_set_row_name(lp, index_contrainte , debut.c_str());
  glp_set_row_bnds(lp, index_contrainte, GLP_LO, varphi.Evalue(x_0).sup() , 0);
  glp_set_mat_row(lp, index_contrainte, index_variable -1, ind_sup, val_sup);
  index_contrainte++;
  
  mtx.unlock();    
}


void vinter::construit_contraintes_hermites( int n , int fin_plus )
{

  // Création du maillage pour mu
  domain total_etat;
  total_etat["x"] = X["x"];
  //  total_etat["x"] = interval(-1,2);
  
    
  mesh une_mesh = mesh(total_etat,n);
  int compteur = 0;
  
  
  mesh::const_iterator i = une_mesh.begin();
  domain x_i = *i;

  // sorte de condition au bord
  double taille_x = (x_i["x"].width()) / fin_plus;
  for (int plus = -fin_plus-1 ; plus < 0 ; plus++)
    {
      domain x_i_bis = *i;
      x_i_bis["x"] = x_i["x"] + interval(plus)*interval(taille_x,taille_x);
      contrainte_hermite(x_i_bis);
    }
  
  
  for (mesh::const_iterator i = une_mesh.begin(); i!=une_mesh.end();++i)
    {
      //      cout << compteur++ << endl;
      domain x_i = *i;

      double taille_x = (x_i["x"].width()) / fin_plus;
      for (int plus = 0 ; plus < fin_plus ; plus++)
	{
	  domain x_i_bis = *i;
	  x_i_bis["x"] = x_i["x"] + interval(plus)*interval(taille_x,taille_x);
	  contrainte_hermite(x_i_bis);
	}
      cout << "-------------" << endl; 
    }
}


// X est le pavé sur lequel la fonction change de valeur
void vinter::contrainte_hermite( domain X )
{

  int nombre_variable = taille*taille + taille;
  int *ind_inf;
  ind_inf = new int[nombre_variable + 1];  
  double *val_inf;
  val_inf = new double[nombre_variable + 1];
  int *ind_sup;
  ind_sup = new int[nombre_variable + 1];  
  double *val_sup;
  val_sup = new double[nombre_variable + 1];
  
  hermite varphi (X["x"]);
  
  /*  string nom_dat;
  nom_dat = "data" +  std::to_string(index_contrainte) + ".dat";
  varphi.to_gnuplot(nom_dat.c_str());*/
  //  X.out();
  
  int index_variable = 1;
  for (mesh::const_iterator i = mu.begin(); i!=mu.end();++i)
    {
      domain x_i = *i;
      
      interval lb_up_psi = interval(-1) * (  varphi.evalue_prime(x_i["x"]) *  f.component["x"].Evalue(x_i));
      ind_inf[index_variable] = index_variable;
      val_inf[index_variable] = lb_up_psi.inf();      

      ind_sup[index_variable] = index_variable;
      val_sup[index_variable] = lb_up_psi.sup();

      index_variable++;
    }

  for (mesh::const_iterator j = nu.begin(); j!=nu.end();++j)
    {
      domain y_j = *j;
      //      temp.out();
      double lb_varphi = varphi.evalue(y_j["x"]).inf();

      ind_inf[index_variable] = index_variable;
      val_inf[index_variable] = lb_varphi;
      //      val_inf[index_variable] = 0;
      

      double up_varphi = varphi.evalue(y_j["x"]).sup();
      ind_sup[index_variable] = index_variable;
      val_sup[index_variable] = up_varphi;
      //      val_sup[index_variable] = 0;

      index_variable++;
    }

  mtx.lock();
  glp_add_rows(lp, 1);
  //  string debut; debut = "contrainte_i_" + nom; 
  //  glp_set_row_name(lp, index_contrainte , debut.c_str());
  glp_set_row_bnds(lp, index_contrainte, GLP_UP, 0, varphi.evalue(x_0["x"]).inf() );
  glp_set_mat_row(lp, index_contrainte, index_variable -1, ind_inf, val_inf);

  /*  for (int index_variable=1 ; index_variable < nombre_variable + 1 ; index_variable++   )
    {
      cout << val_inf[index_variable] << ";";
    }
  cout << endl;

  for (int index_variable=1 ; index_variable < nombre_variable + 1 ; index_variable++   )
    {
      cout << val_sup[index_variable] << ";";
    }
  cout << endl;
  */

  
  index_contrainte++;

  glp_add_rows(lp, 1);
  //  debut = "contrainte_s_" + nom;
  //  glp_set_row_name(lp, index_contrainte , debut.c_str());
  glp_set_row_bnds(lp, index_contrainte, GLP_LO, varphi.evalue(x_0["x"]).sup() , 0);
  glp_set_mat_row(lp, index_contrainte, index_variable -1, ind_sup, val_sup);
  index_contrainte++;
  
  mtx.unlock();    
}





// fonction qui résoud le problème primal 
double vinter::primal()
{};




