#ifndef __VINTER__
#define __VINTER__

#include <mutex>         
#include <glpk.h>
//---------------------------------------------------------------------------
#include "expression.h"
#include "domain.h"
#include "map.h"
#include "mesh.h"
#include "iboolean.h"

typedef filib::interval<double,filib::native_switched,filib::i_mode_extended> interval;

class vinter
{

 private :
  mesh mu; // ceci est un découpage on ne peut plus parfait de l'espace [0,T] x X x U
  mesh nu; // ceci est un découpage l'espace {T} x X 
  
  int taille;

  glp_prob *lp;

  int index_a;
  int index_contrainte;

 public :
  
  int cardinal_variable();
  int cardinal_contraints();
  
  Map f; //code le champs de vecteurs qui dépend de t,x,u
  
  expression h; //cout infinitésimal qui dépend de t,x,u aussi appelé le running cost
  expression H; //cout final défini sur X \times {t_final}

  domain K; //domaine "cible" où doit finir x(T), dans la première version K est simplement X
  domain T; //domaine temporelle de la forme [0,t_final]
  domain X; //domaine des états dans lequel doit rester l'état
            //on pourra à l'avenir rajouter des équations qui définissent X
  domain U; //domaine des commandes dans lequel doit rester la commande 
            //on pourra à l'avenir rajouter des équations qui définissent U

  domain x_0;

  vinter (Map _f  , expression _h, expression _H, domain _K  , domain _T, domain _X , domain _U , domain x_0);
  
  void construit_mesh(int nb);
  void construit_cout();  
  string contraintes_temps_total();
  string contraintes_inclusion();
  string dirac_final();
  void construit_contraintes(int debut, int fin);  


  // créer une contrainte avec une fonction polynome
  void contrainte(string polynome, string nom);

  // créer une contrainte avec une fonction de la classe hermite dont le support est X
  void contrainte_hermite( domain X );

  // créer des contraintes hermites partir de la mesh qui est découper en n dans chaque dimension
  void construit_contraintes_hermites( int n , int fin );

  
  void contraintes_temporelles();

  // fonction qui résoud 
  double solve(string);
  void solution_to_gnuplot(char * );
  void show_result();
  
  
  // fonction qui résoud le problème primal 
  double primal(); 

  // truc pour générer les contraintes en parallèle
  mutex mtx;         

  // code qui permet de connaitre la borne inférieure accessible avec le maillaure donnée
  double test_vrai_solution();
  void write();
};


//---------------------------------------------------------------------------
#endif
